# encoding: utf-8
from __future__ import unicode_literals
from __future__ import absolute_import

from lxml.etree import tostring
from collections import defaultdict

from django.db.models import Max

from kinder.core.children.models import (
    ChildrenDelegate,
    Delegate,
    DelegateTypeEnumerate,
)
from kinder.core.group.enum import GroupStatusEnum
from kinder.core.direct.models import Direct
from kinder.core.declaration.models import (
    Declaration as DeclarationModel,
    DeclarationUnit,
)
from kinder.core.declaration.declaration_history.models import (
    DeclarationHistory,
)
from kinder.core.declaration_status.enum import DSS
from kinder.core.group.enum import GroupStatusEnum
from kinder.core.group.models import Pupil

from spyne.util.xml import get_object_as_xml

from .types import (
    Declaration,
    Parent,
    Root,
)

from ..utils import AddressBlock
from ..constants import DECLARATION_HISTORY_TABLES


def get_parents(children_ids):
    """Получение информации о родителях детей указанных в заявлениях."""
    # Родителем считается здесь только законный представитель
    parent_data = ChildrenDelegate.objects.filter(
        children_id__in=children_ids,
        delegate__type=DelegateTypeEnumerate.LEX,
    ).values(
        'children_id',
        'delegate_id',
        'delegate__snils',
        # addressResid
        'delegate__reg_address_place',
        'delegate__reg_address_street',
        'delegate__reg_address_house',
        # addressStay
        'delegate__delegatecontingent__stay_address_place',
        'delegate__delegatecontingent__stay_address_street',
        'delegate__delegatecontingent__stay_address_house',
        'delegate__delegatecontingent__stay_address_corps',
        # addressFact
        'delegate__address_place',
        'delegate__address_street',
        'delegate__address_house',
        'delegate__address_corps',
    )
    parents = defaultdict(list)
    for pd in parent_data:
        addr = AddressBlock('delegate', pd, 'delegatecontingent')
        parents[pd['children_id']] = [
            Parent(
                parentID=pd['delegate_id'],
                parentSNILS=bool(pd['delegate__snils']),
                addressResid=addr.reg,
                addressStay=addr.stay,
                addressFact=addr.fact,
            )]
    return parents


def get_directs(declaration_ids):
    """Получение информации о направлениях связанных с заявлениями."""
    # По бизнес-логике на одно заявление
    # идет одно направление
    direct_data = Direct.objects.filter(
        declaration_id__in=declaration_ids,
        temporary=False,
    ).exclude(
        group__status=GroupStatusEnum.ARCHIVE
    ).values(
        'declaration_id',
        'group__health_need',
    )
    directs = defaultdict(dict)
    for dd in direct_data:
        dd['declaration_id'] = {
            'health_need': dd['group__health_need'],
        }
    return directs


def get_units(declaration_ids):
    """Получение желаемых учреждений."""
    unit_data = DeclarationUnit.objects.filter(
        declaration_id__in=declaration_ids
    ).values_list(
        'declaration_id', 'unit_id', 'ord')
    units = defaultdict(dict)
    for decl_id, unit_id, pos in unit_data:
        units[decl_id][pos] = unit_id
    return units


def get_current_units(children_ids):
    """Получение текущих учреждений из зачислений.

    Берется одно последнее по дате зачисление по каждому ребенку,
    за исключением архивных групп.
    """
    pupil_queryset = Pupil.objects.exclude(
        grup__status__in=GroupStatusEnum.ARCHIVE
    ).filter(
        children__in=children_ids
    ).values('id', 'children_id', 'date', 'grup__unit')
    latest_pupils = pupil_queryset.values_list(
        'children_id').annotate(latest_date=Max('date'))
    current_units = defaultdict(dict)
    for children_id, latest_date in latest_pupils:
        current_units[children_id] = pupil_queryset.filter(
            children_id=children_id, date=latest_date
        ).values_list('grup__unit_id', flat=True).first()
    return current_units


def get_changes(declaration_ids):
    """Получение последних изменений по заявлениям."""
    latest_changes = DeclarationHistory.objects.filter(
        declaration_id__in=declaration_ids,
    ).values('declaration_id').annotate(change_date=Max('date'))
    return {i['declaration_id']: i['change_date'] for i in latest_changes}


def get_content(ctx, report_code, decl_date):
    """Точка входа метода, возвращает строку xml."""
    # Финализация кверисета иначе далее проблемы
    changed_or_created = set(DeclarationHistory.objects.filter(
        date__year=decl_date.year,
        date__month=decl_date.month,
        date__day=decl_date.day,
        table__in=DECLARATION_HISTORY_TABLES,
    ).values_list('declaration_id', flat=True))

    # т.к. DeclarationHistory была введена не сразу, необходимо также добавить
    # также все заявления с указанной датой подачи
    changed_or_created.update(
        DeclarationModel.objects.filter(
            date__year=decl_date.year,
            date__month=decl_date.month,
            date__day=decl_date.day,
        ).values_list('id', flat=True)
    )

    declaration_data = DeclarationModel.objects.filter(
        id__in=changed_or_created
    ).exclude(
        status__code=DSS.ARCHIVE
    ).values(
        'id',
        'client_id',  # identificator
        'date',  # creationDate
        'source',
        'desired_date',  # prefDate
        'best_privilege_id',  # privID
        'status__code',  # douChange
        'consent_short_time_group',

        'children_id',
        'children__snils',  # childSNILS
        'children__health_need_id',  # healthProblem
        'children__date_of_birth',  # birthdate

        # addressResid
        'children__reg_address_place',
        'children__reg_address_street',
        'children__reg_address_house',
        # addressStay
        'children__childrencontingent__stay_address_place',
        'children__childrencontingent__stay_address_street',
        'children__childrencontingent__stay_address_house',
        'children__childrencontingent__stay_address_corps',
        # addressFact
        'children__address_place',
        'children__address_street',
        'children__address_house',
        'children__address_corps',
    )

    directs = get_directs(changed_or_created)
    units = get_units(changed_or_created)
    children_ids = set(dd['children_id'] for dd in declaration_data)
    current_units = get_current_units(children_ids)
    parents = get_parents(children_ids)
    changes = get_changes(changed_or_created)

    declarations = []
    for decl in declaration_data:
        want_change_dou = decl['status__code'] == DSS.WANT_CHANGE_DOU
        addr = AddressBlock('children', decl, 'childrencontingent')
        declarations.append(Declaration(
            id=decl['id'],
            identificator=decl['client_id'],
            creationDate=decl['date'],
            changeDate=changes.get(decl['id']),
            sourceID=decl['source'],
            birthdate=decl['children__date_of_birth'],
            prefDate=decl['desired_date'],
            firstID=units[decl['id']].get(1),
            secondID=units[decl['id']].get(2),
            thirdID=units[decl['id']].get(3),
            privID=decl['best_privilege_id'],
            shottermGroup=decl['consent_short_time_group'],
            douChange=want_change_dou,
            currentID=(want_change_dou and current_units[decl['children_id']]
                       or None),
            addressResid=addr.reg,
            addressStay=addr.stay,
            addressFact=addr.fact,
            parents=parents[decl['children_id']],
            childSNILS=bool(decl['children__snils']),
            healthGroup=directs[decl['id']].get('health_need'),
            healthProblem=decl['children__health_need_id'],
        ))

    root = Root(declarations=declarations)
    result = tostring(
        get_object_as_xml(root), pretty_print=True, encoding='UTF-8')
    return result
