# encoding: utf-8
from __future__ import unicode_literals
from __future__ import absolute_import

import json
from itertools import groupby
from datetime import datetime, time
from lxml.etree import tostring
from collections import defaultdict, namedtuple

from kinder.core.declaration.declaration_history.models import DeclarationHistory
from kinder.core.declaration.models import DeclarationStatusLog

from kinder.users.models import User

from spyne.util.xml import get_object_as_xml

from ..constants import DECLARATION_HISTORY_TABLES

from .types import (
    ChangesList,
    StatusChange,
    Root,
)

from .constants import (
    LOG_FIELDS_TO_CHECK,

    DECLARATION_MO,
    DECLARATION_OFFER_OTHER,
    DECLARATION_DATE,
    DECLARATION_BEST_PRIVILEGE,
    DECLARATION_DESIRED_DATE,
    DECLARATION_UNIT_FIELDS,
    DECLARATION_CARE_GROUP,
    DECLARATION_SHORT_TIME_GROUP,

    CHILDREN_SNILS,
    CHILDREN_HEALTH_NEED,
    CHILDREN_BIRTH_DATE,
    CHILDREN_FULLNAME,
    CHILDREN_DOCUMENT_FIELDS,
    CHILDREN_ADDRESS_FIELDS,

    CONTINGENT_HEALTH_GROUP,
)

Period = namedtuple('Period', 'start, end')


def get_time_period(dates, end_date):
    """
    Возвращает период с прошлой смены статуса по настоящую.

    :param dates: Даты изменений статусов за день по заявлению
    :param end_date: Дата смены статуса, для рассчета
    """
    # В день произошла только одна смена статуса у заявления
    # в таком случае отдаем период как день целиком
    if len(dates) in (0, 1):
        period = Period(
            datetime.combine(end_date, time.min),
            datetime.combine(end_date, time.max))
    else:
        # Получаем предыдущую дату из списка
        date_index = dates.index(end_date)
        start_date = (
            dates[date_index - 1] if date_index else
            # Если date_index == 0, т.е. елемент первый,
            # то дата начала периода это начало дня
            datetime.combine(end_date, time.min))
        period = Period(start_date, end_date)
    return period


def get_changes_list(period, fields):
    """
    Заполняет список измененных полей.

    :param period: Рассчетный период для выборки полей
    :param fields: Список измененных полей за день, вместе с датами
    """
    if not fields:
        return []

    # Выборка полей которые были изменены в период
    # между сменой статуса заявления
    fields = set(
        field for field, date in fields
        if period.start <= date <= period.end)

    other = fields.difference(LOG_FIELDS_TO_CHECK)
    return ChangesList(
        creationDate=DECLARATION_DATE in fields,
        prefDate=DECLARATION_DESIRED_DATE in fields,
        fullname=CHILDREN_FULLNAME in fields,
        birthdate=CHILDREN_BIRTH_DATE in fields,
        privilege=DECLARATION_BEST_PRIVILEGE in fields,
        healthGroup=CONTINGENT_HEALTH_GROUP in fields,
        healthProblem=CHILDREN_HEALTH_NEED in fields,
        prefDou=bool(fields.intersection(DECLARATION_UNIT_FIELDS)),
        mo=DECLARATION_MO in fields,
        document=bool(fields.intersection(CHILDREN_DOCUMENT_FIELDS)),
        SNILS=CHILDREN_SNILS in fields,
        otherVariants=DECLARATION_OFFER_OTHER in fields,
        accceptCareGroup=DECLARATION_CARE_GROUP in fields,
        acceptShorttermGroup=DECLARATION_SHORT_TIME_GROUP in fields,
        address=bool(fields.intersection(CHILDREN_ADDRESS_FIELDS)),
        other=bool(other),
    )


def get_changed_fields(changed_declarations, decl_date):
    """
    Получение из логов заявлений полей в которых отмечены изменения.

    :param changed_declarations: Кортеж id заявлений для выборки
    :param decl_date: Рассчетная дата из запроса

    returns: {<declaration_id>: (('mo', datetime), ('status_id', datetime)), ...}
    """
    changed_fields = DeclarationHistory.objects.filter(
        declaration_id__in=changed_declarations,
        date__year=decl_date.year,
        date__month=decl_date.month,
        date__day=decl_date.day,
        table__in=DECLARATION_HISTORY_TABLES,
    ).values_list(
        'declaration_id',
        'field',
        'table',
        'date',
    )

    result = defaultdict(tuple)
    for decl_id, field, table, date in changed_fields:
        # При удалении ЖУ поле field будет пустым
        if table == 'declaration_unit' and field is None:
            field = 'remove_declaration_unit'
        result[decl_id] = result[decl_id] + ((field, date), )
    return result


def get_content(ctx, report_code, decl_date):
    """Точка входа метода, возвращает строку xml."""

    # Отношение пользователь - учреждение выбранное администратором
    users_units = dict(
        User.objects.values_list('id', 'userprofile__max_unit'))

    # Записи о изменении статусов заявлений
    changes = tuple(DeclarationStatusLog.objects.filter(
        datetime__year=decl_date.year,
        datetime__month=decl_date.month,
        datetime__day=decl_date.day,
    # Сортировка далее используется в groupby
    ).order_by(
        'declaration_id',
    ).values(
        'id',
        'declaration_id',
        'datetime',
        'user_id',
        'status_id',
    ))

    # Преобразовает данные в такую структуру:
    # {<declaration_id>: (datetime, datetime, ...), ...}
    # карта id заявления - отсортированный кортеж дат смены статусов
    changed_declarations = defaultdict(tuple, {
        decl_id: tuple(sorted(i['datetime'] for i in items))
        for decl_id, items in
        groupby(changes, key=lambda c: c['declaration_id'])})

    # Измененные поля, передаем id заявлений у которых был изменен статус
    # в рассчетную дату и саму рассчетную дату
    changed_fields = get_changed_fields(changed_declarations.keys(), decl_date)

    status_changes = []
    for c in changes:
        # Рассчетный период для получения списка измененных полей
        period = get_time_period(
            changed_declarations[c['declaration_id']], c['datetime'])
        status_changes.append(StatusChange(
            id=c['id'],
            declarationID=c['declaration_id'],
            changeDate=c['datetime'],
            statusID=c['status_id'],
            douID=users_units.get(c['user_id']),
            changesList=get_changes_list(
                period=period,
                fields=changed_fields[c['declaration_id']])
        ))

    root = Root(statusChanges=status_changes)
    result = tostring(
        get_object_as_xml(root), pretty_print=True, encoding='UTF-8')
    return result
