# encoding: utf-8
from __future__ import unicode_literals
from __future__ import absolute_import

from collections import defaultdict

from spyne import Fault

from m3_fias.utils import get_address_object, get_address_object_name
from kinder.core.utils.fias import fias_field_validator

from .constants import DECL_INFO_REPORT_TYPES
from .types import Address


def validate_dou_info_request(ctx, decl_date, report_code):
    u"""Проверяет входные данные."""
    error = None

    if not report_code in DECL_INFO_REPORT_TYPES:
        error = u'Задан невалидный тип отчета'

    if error is not None:
        raise Fault(faultstring=error)


class AddressBlock(object):
    """Формирует блоки адресов - по регистрации, факт. и проживания."""

    def __init__(self, model_name, model_values, contingent_model_name=None):
        self.model_name = model_name.lower()
        self.model_values = model_values
        self.contingent_model_name = contingent_model_name

    SCOPE = {
        'locality': 'place',
        'street': 'street',
        'house': 'house',
        'building': 'corps',
    }

    FACT_ADDRESS = 'fact'
    STAY_ADDRESS = 'stay'
    REG_ADDRESS = 'reg'
    JUR_ADDRESS = 'jur'
    ADDRESS_TYPES = (
        FACT_ADDRESS,
        STAY_ADDRESS,
        REG_ADDRESS,
        JUR_ADDRESS,
    )

    def __getattr__(self, address_type):
        if address_type not in self.ADDRESS_TYPES:
            return object.__getattribute__(self, address_type)
        return Address(**self._address_values(address_type))

    @staticmethod
    def _decode(address):
        # Получение списка полей в которых есть коды фиас
        if not any(address.values()):
            return address
        fias_fields = tuple(
            name for name, value in address.items()
            if value and fias_field_validator.regex.match(value))
        # Вообще номер дома хранится в двух полях
        if not fias_fields:
            return address
        # Попытка получить адрес до улицы
        if address['street']:
            fias_data = get_address_object(address['street'])
            # Заполняем адрес и населенный пункт по коду фиас
            if fias_data:
                address['street'] = get_address_object_name(
                    get_address_object(fias_data.guid))
                # Получаем населенный пункт из полного адреса
                address['locality'] = get_address_object_name(
                    get_address_object(fias_data.parent_guid))
        else:
            # Попытка получить адрес до населенного пункта
            if address['locality']:
                fias_data = get_address_object(address['locality'])
                if fias_data:
                    address['locality'] = get_address_object_name(
                        get_address_object(fias_data.guid))
        return address

    def _address_field(self, address_type, scope):
        """Формирует ключ для получения адреса."""
        prefix = (
            '%s_' % address_type if address_type != self.FACT_ADDRESS else '')
        is_contingent = address_type in (self.STAY_ADDRESS, self.JUR_ADDRESS)
        if is_contingent and self.contingent_model_name:
            prefix = '%s__%s' % (self.contingent_model_name, prefix)
        return '%saddress_%s' % (prefix, scope)

    def _address_values(self, address_type):
        """Возвращает генератор кортежей - ключевой аргумент и адрес."""
        # Цикл по масштабу адреса - город, улица, дом, корпус
        address = defaultdict(str)
        for ias_to_s, kinder_s in self.SCOPE.items():
            address_field = self._address_field(address_type, kinder_s)
            # Составляется полный ключ запроса
            # например children__childrencontingent__reg_address__street
            field_name = '{model}__{field}'.format(
                model=self.model_name, field=address_field)
            # Пытаемся взять значение по ключу children__address_street
            # Но если не получилось попытаемся еще взять значение напрямую
            # с ключем address_street. Не всегда адрес нужно брать через отношение
            address[ias_to_s] = self.model_values.get(
                field_name, self.model_values.get(address_field))

        return self._decode(address)
